/* YaRock Player
* Copyright (C) 2010-2013 Sebastien amardeilh (sebastien.amardeilh+yarock@gmail.com)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "playlist_parser.h"
#include "debug.h"

#include <QtCore>

/*******************************************************************************
    Extension
*******************************************************************************/
const QStringList m3u_extension  = QStringList() << "m3u" << "m3u8";
const QStringList pls_extension  = QStringList() << "pls";
const QStringList xspf_extension = QStringList() << "xspf";

/*******************************************************************************
    Internals functions
*******************************************************************************/
QList<MediaItemPtr>  readM3uPlaylist(QIODevice* device, const QDir& playlist_dir=QDir() );
QList<MediaItemPtr>  readPlsPlaylist(QIODevice* device, const QDir& playlist_dir=QDir() );
QList<MediaItemPtr>  readXspfPlaylist(QIODevice* device, const QDir& playlist_dir=QDir() );

void saveM3uPlaylist(QIODevice* device, const QDir& playlist_dir, QList<MediaItemPtr> list);
void savePlsPlaylist(QIODevice* device, const QDir& playlist_dir, QList<MediaItemPtr> list);
void saveXspfPlaylist(QIODevice* device, const QDir& playlist_dir, QList<MediaItemPtr> list);

/*
********************************************************************************
*                                                                              *
*    PlaylistFromFile                                                          *
*                                                                              *
********************************************************************************
*/
QList<MediaItemPtr> MEDIA::PlaylistFromFile(const QString& filename)
{
    Debug::debug() << "MEDIA::PlaylistFromFile";
    QList<MediaItemPtr>   list;
        
    // Open the file
    QFileInfo info(filename);
    QFile file(filename);


    if (!file.open(QFile::ReadOnly)) {
      Debug::debug() << "Failed to open read only file " << filename;
      return list;
    }
    
    QString extension = info.suffix().toLower();
    QDir playlist_dir = QDir(info.absoluteDir());
    
    // choose parser
    if( m3u_extension.contains(extension) ) 
      list = readM3uPlaylist( &file, playlist_dir);
    else if( pls_extension.contains(extension) )
      list = readPlsPlaylist( &file, playlist_dir);
    else if( xspf_extension.contains(extension ))
      list = readXspfPlaylist( &file, playlist_dir );
    
    return list;
}

/*
********************************************************************************
*                                                                              *
*    PlaylistFromBytes                                                         *
*                                                                              *
********************************************************************************
*/
QList<MediaItemPtr> MEDIA::PlaylistFromBytes(QByteArray& bytes)
{
    Debug::debug() << "MEDIA::PlaylistFromBytes";
    QList<MediaItemPtr>   list;    
    
    QBuffer buffer(&bytes);
    if (!buffer.open(QFile::ReadOnly)) {
      Debug::debug() << "Failed to open data";
      return list;
    }
    
    // choose parser
    QByteArray peeked_data = buffer.peek(512);
    if( peeked_data.contains("#EXTM3U") || peeked_data.contains("#EXTINF") ) 
      list = readM3uPlaylist( &buffer ) ;
    else if( peeked_data.toLower().contains("[playlist]") )
      list = readPlsPlaylist( &buffer ) ;
    else if( peeked_data.contains("<playlist") && peeked_data.contains("<trackList") )
      list = readXspfPlaylist( &buffer ) ;
    else // try m3u style
      list = readM3uPlaylist( &buffer ) ;
    
    return list;
}    
    

/*******************************************************************************
    M3U playlist read
*******************************************************************************/
QList<MediaItemPtr>  readM3uPlaylist(QIODevice* device, const QDir& playlist_dir )
{
    QList<MediaItemPtr>   list;

    QString data = QString::fromUtf8(device->readAll());
  
    if(data.isEmpty()) return list;
    data.replace('\r', '\n');
    data.replace("\n\n", "\n");

    QByteArray bytes = data.toUtf8();
    QBuffer buffer(&bytes);
    buffer.open(QIODevice::ReadOnly);
  
    //! metadata
    int       lenght;
    QString   title;
    QString   artist;
    
    //! main reading loop
    QString line;
    
    while(!buffer.atEnd()) {
      line = QString::fromUtf8(buffer.readLine()).trimmed();

      if(line.startsWith("#EXT")) 
      {
        QString info = line.section(':', 1);
        QString l = info.section(',', 0, 0);
        bool ok = false;
        int length = l.toInt(&ok);
        if (!ok) {lenght = -1; continue;}

        QString track_info = info.section(',', 1);
        QStringList list = track_info.split('-');

        if (list.size() <= 1) {
          title   = track_info;
          continue;
        }

        artist = list[0].trimmed();
        title = list[1].trimmed();
      }
      else if (line.startsWith('#')) 
      {
        continue;
      }
      else if( !line.isEmpty() )
      {
        //Debug::debug() << "---- readM3uPlaylist -> line " << line << "\n";
        MediaItemPtr mi = MediaItemPtr(new MediaItem());

        //! Find the Track location
        if (line.contains(QRegExp("^[a-z]+://"))) {
          QUrl url(line);
          if (url.isValid()) {
              mi->type                      = TYPE_STREAM;
              mi->data[STREAM::id]          = -1;
              mi->data[STREAM::url]         = url.toString();
              mi->data[STREAM::name]        = title.isEmpty() ? url.toString() : title;
              mi->data[STREAM::isFavorite]  = false;
              mi->data[STREAM::isPlaying]   = false;
              mi->data[STREAM::isBroken]    = false;
              mi->data[STREAM::isPlayed]    = false;
              mi->data[STREAM::isStopAfter] = false;

              list.append(mi);
          }
        }        
        else {

          QString file_path = line;

          file_path = QDir::fromNativeSeparators(file_path);
          
          // Make the path absolute
          if (!QDir::isAbsolutePath(file_path))
            file_path = playlist_dir.absoluteFilePath(file_path);

          // Use the canonical path
          if (QFile::exists(file_path))
            file_path = QFileInfo(file_path).canonicalFilePath();

          mi->type                     = TYPE_TRACK;
          mi->data[TRACK::id]          =  -1;
          mi->data[TRACK::url]         =  file_path;
          mi->data[TRACK::title]       =  title.isEmpty() ? QString() : title;
          mi->data[TRACK::duration]    =  (lenght!=-1) ? lenght : 0;
          mi->data[TRACK::artist]      =  artist.isEmpty() ? QString() : artist;
          mi->data[TRACK::isPlaying]   =  false;
          mi->data[TRACK::isBroken]    =  QFile::exists(file_path) ? false : true;
          mi->data[TRACK::isPlayed]    =  false;
          mi->data[TRACK::isStopAfter] =  false;

          list.append(mi);
  
        }


        lenght   = -1;
        title    = "";
        artist   = "";
      }
    } // end while
    
  return list;
}

/*******************************************************************************
    .pls playlist read
*******************************************************************************/
QList<MediaItemPtr>  readPlsPlaylist(QIODevice* device, const QDir& playlist_dir )
{
    QList<MediaItemPtr> list;

    MediaItemPtr mi = MediaItemPtr(0);
    
    while (!device->atEnd()) {
      QString line = QString::fromUtf8(device->readLine()).trimmed();
      int equals = line.indexOf('=');
      QString key = line.left(equals).toLower();
      QString value = line.mid(equals + 1);

      if (key.startsWith("file")) 
      {
        mi = MediaItemPtr(new MediaItem());
        list.append(mi);

        //! Find the Track location
        if (value.contains(QRegExp("^[a-z]+://"))) {
          QUrl url(value);
          if (url.isValid()) {
              // Debug::debug() << "---- readPlsPlaylist -> url.isValid()" << url;
              mi->type                      = TYPE_STREAM;
              mi->data[STREAM::id]          = -1;
              mi->data[STREAM::url]         = value;
              mi->data[STREAM::name]        = value;
              mi->data[STREAM::isFavorite]  = false;
              mi->data[STREAM::isPlaying]   = false;
              mi->data[STREAM::isBroken]    = false;
              mi->data[STREAM::isPlayed]    = false;
              mi->data[STREAM::isStopAfter] = false;            
          }
        }
        else {
          QString file_path = value;

          file_path = QDir::fromNativeSeparators(file_path);
          
          // Make the path absolute
          if (!QDir::isAbsolutePath(file_path))
            file_path = playlist_dir.absoluteFilePath(file_path);

          // Use the canonical path
          if (QFile::exists(file_path))
            file_path = QFileInfo(file_path).canonicalFilePath();

          mi->type                     = TYPE_TRACK;
          mi->data[TRACK::id]          =  -1;
          mi->data[TRACK::url]         =  file_path;
          mi->data[TRACK::isPlaying]   =  false;
          mi->data[TRACK::isBroken]    =  QFile::exists(file_path) ? false : true;
          mi->data[TRACK::isPlayed]    =  false;
          mi->data[TRACK::isStopAfter] =  false;          
        }    
      } // key is filename
      else if (key.startsWith("title")) 
      {
        if(mi->type == TYPE_TRACK) 
          mi->data[TRACK::title] = value;
        else 
          mi->data[STREAM::name] = value;
      } 
      else if (key.startsWith("length"))
      {
        if(mi->type == TYPE_TRACK) 
          mi->data[TRACK::duration] = value.toInt();
      }
    } // fin while
    
    
    return list;
}  


/*******************************************************************************
    .xspf playlist read
*******************************************************************************/
QList<MediaItemPtr>  readXspfPlaylist(QIODevice* device, const QDir& playlist_dir )
{
    QList<MediaItemPtr>  list;

    QXmlStreamReader xml(device);
  
    MediaItemPtr mi;
    
    while (!xml.atEnd()) {
      QXmlStreamReader::TokenType type = xml.readNext();

      if ((type == QXmlStreamReader::StartElement) && (xml.name() == "track")) 
      {
        mi = MediaItemPtr(new MediaItem());
      }
      else if ((type == QXmlStreamReader::StartElement) && (xml.name() == "location")) 
      {
            QString file_path = QString(xml.readElementText());

            //Debug::debug() << "---- readXspfPlaylist -> Find the Track location" << file_path;
            if (file_path.contains(QRegExp("^[a-z]+://"))) {
              QUrl url(file_path);
              if (url.isValid()) {

                //Debug::debug() << "---- readXspfPlaylist -> url.isValid()" << url;
                mi->type                      = TYPE_STREAM;
                mi->data[STREAM::id]          = -1;
                mi->data[STREAM::url]         = file_path;
                mi->data[STREAM::name]        = file_path;
                mi->data[STREAM::isFavorite]  = false;
                mi->data[STREAM::isPlaying]   = false;
                mi->data[STREAM::isBroken]    = false;
                mi->data[STREAM::isPlayed]    = false;
                mi->data[STREAM::isStopAfter] = false;
              }
            }
            else {
              file_path = QDir::fromNativeSeparators(file_path);
              
              // Make the path absolute
              if (!QDir::isAbsolutePath(file_path))
                file_path = playlist_dir.absoluteFilePath(file_path);
            
              // Use the canonical path
              if (QFile::exists(file_path))
                file_path = QFileInfo(file_path).canonicalFilePath();
              
              mi->type                     = TYPE_TRACK;
              mi->data[TRACK::id]          =  -1;
              mi->data[TRACK::url]         =  file_path;
              mi->data[TRACK::isPlaying]   =  false;
              mi->data[TRACK::isBroken]    =  QFile::exists(file_path) ? false : true;
              mi->data[TRACK::isPlayed]    =  false;
              mi->data[TRACK::isStopAfter] =  false;
           }
      } // end location
      else if ((type == QXmlStreamReader::StartElement) && (xml.name() == "title")) 
      {
          if(mi->type == TYPE_TRACK) 
            mi->data[TRACK::title] = QString(xml.readElementText());
          else 
            mi->data[STREAM::name] = QString(xml.readElementText());
      } 
      else if ((type == QXmlStreamReader::EndElement) && (xml.name() == "track")) 
      {
        //Debug::debug() << "---- readXspfPlaylist -> list.append(mi)" << mi;
        if(mi) 
          list.append(mi);
        mi = MediaItemPtr(0);
      }
    }  // End while xml end

    return list;
}  



/*
********************************************************************************
*                                                                              *
*    PlaylistToFile                                                            *
*                                                                              *
********************************************************************************
*/
void MEDIA::PlaylistToFile(const QString& filename, QList<MediaItemPtr> list) 
{
    // Open the file
    QFileInfo info(filename);

    QFile file(filename);
    if (!file.open(QFile::WriteOnly)) {
      Debug::debug() << "Failed to open file for writing" << filename;
      return;
    }
    
    QString extension = info.suffix().toLower();
    QDir playlist_dir = QDir(info.absoluteDir());
        
    if( m3u_extension.contains(extension) )
      saveM3uPlaylist(&file, playlist_dir, list);
    else if( pls_extension.contains(extension) )
      savePlsPlaylist(&file, playlist_dir, list);
    else if( xspf_extension.contains(extension) )
      saveXspfPlaylist(&file, playlist_dir, list);
      
    file.close();
}

/*******************************************************************************
    m3u playlist save
*******************************************************************************/
void saveM3uPlaylist(QIODevice* device, const QDir& playlist_dir, QList<MediaItemPtr> list) 
{
    device->write("#EXTM3U\n");

    foreach (MediaItemPtr media, list) 
    {
      if (!media) continue;
      if (media->type != TYPE_TRACK && media->type != TYPE_STREAM) continue;

      QString info;
      QString media_path; 
      if (media->type == TYPE_TRACK ) {
        info = QString("#EXTINF:%1,%2 - %3\n").arg(media->data.at(TRACK::duration).toString())
                                              .arg(media->data.at(TRACK::artist).toString())
                                              .arg(media->data.at(TRACK::album).toString());
        device->write(info.toUtf8());
    
        //! Get the path to MediaItem relative to the Playlist File directory
        media_path = media->data.at(TRACK::url).toString();
        media_path = playlist_dir.relativeFilePath(media_path);
      }
      else {
        info = QString("#EXTINF:\n");
        info = QString("#EXTINF:%1,%2 - %3\n").arg(0)
                                              .arg("")
                                              .arg(media->data.at(STREAM::name).toString());
       
        device->write(info.toUtf8());
        media_path = media->data.at(STREAM::url).toString();
      }
     
    device->write(media_path.toUtf8());
    device->write("\n");
    }
}


/*******************************************************************************
    .pls playlist save
*******************************************************************************/
void savePlsPlaylist(QIODevice* device, const QDir& playlist_dir, QList<MediaItemPtr> list) 
{
    QTextStream stream(device);
    stream << "[playlist]" << endl;
    stream << "Version=2" << endl;
    stream << "NumberOfEntries=" << list.size() << endl;

    int n = 1;
    foreach (MediaItemPtr media, list) 
    {
      if (!media) continue;
      if (media->type != TYPE_TRACK && media->type != TYPE_STREAM) continue;
        
      // write metadata
      if (media->type == TYPE_TRACK ) {
        QString media_path = media->data.at(TRACK::url).toString();
        media_path = playlist_dir.relativeFilePath(media_path).toUtf8();
        
        stream << "File" << n << "=" << media_path << endl;
        stream << "Title" << n << "=" << media->data.at(TRACK::title).toString() << endl;
        stream << "Length" << n << "=" << media->data.at(TRACK::duration).toInt()  << endl;
        ++n;
      }
      else { // TYPE_STREAM
        stream << "File" << n << "=" << media->data.at(STREAM::url).toString().toUtf8() << endl;
        stream << "Title" << n << "=" << media->data.at(STREAM::name).toString().toUtf8() << endl;
        stream << "Length" << n << "=" << endl;
        ++n;
      }
    } // fin Foreach
}


/*******************************************************************************
    .xspf playlist save
*******************************************************************************/
void saveXspfPlaylist(QIODevice* device, const QDir& playlist_dir, QList<MediaItemPtr> list) 
{
    QXmlStreamWriter xml(device);
    xml.writeStartDocument();
    xml.writeStartElement("playlist");
    xml.writeAttribute("version", "1");
    xml.writeDefaultNamespace("http://xspf.org/ns/0/");

    xml.writeStartElement("trackList");
    foreach (MediaItemPtr media, list) 
    {
      if (!media) continue;
      if (media->type != TYPE_TRACK && media->type != TYPE_STREAM) continue;
          
      xml.writeStartElement("track");
    
      QString media_path;
      // write metadata
      if (media->type == TYPE_TRACK ) {
        media_path = media->data.at(TRACK::url).toString();
        media_path = playlist_dir.relativeFilePath(media_path).toUtf8();
        
        xml.writeTextElement("location", media_path);        
        xml.writeTextElement("title", media->data.at(TRACK::title).toString());
        xml.writeTextElement("creator", media->data.at(TRACK::artist).toString());
        xml.writeTextElement("album", media->data.at(TRACK::album).toString());
        xml.writeTextElement("duration", QString::number(media->data.at(TRACK::duration).toInt() * 1000));
      }
      else { // TYPE_STREAM
        media_path = media->data.at(STREAM::url).toString();

        xml.writeTextElement("location", media_path);        
        xml.writeTextElement("title", media->data.at(STREAM::name).toString());
      }
    
      xml.writeEndElement(); //track
    }    
    xml.writeEndElement(); //trackList
    xml.writeEndElement(); //playlist
    xml.writeEndDocument(); 
}
