/* YaRock Player
* Copyright (C) 2010-2013 Sebastien amardeilh (sebastien.amardeilh+yarock@gmail.com)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
 
#include "Playlist/playlisteditor.h"
#include "Playlist/playlistmodel.h"
#include "Playlist/playlistview.h"

#include "threadmanager.h"
#include "statuswidget.h"
#include "Dialog/addstreamdialog.h"
#include "debug.h"

#include <QtGui>
#include <QtCore>

/*
********************************************************************************
*                                                                              *
*    Class PlaylistEditor                                                      *
*                                                                              *
********************************************************************************
*/
PlaylistEditor::PlaylistEditor(QWidget *parent) : QDialog(parent)
{
    //! init
    model_ = new PlaylistModel(this);
    view_  = new PlaylistView(this, model_);
    view_->setMode((int) PLAYQUEUE::MODE_ARTIST);   
    
    m_thread_manager  = new ThreadManager();
    m_thread_manager->setModel(model_);
    model_->setThreadManager(m_thread_manager);   
    
    QObject::connect(m_thread_manager,SIGNAL(playlistSaved()),this,SIGNAL(playlistSaved()));
    QObject::connect(m_thread_manager,SIGNAL(playlistSaved()),this,SIGNAL(playlistSaved()));
    
    // GUI
    create_ui();
    
    QObject::connect(ui_line_edit,SIGNAL(textfield_changed()),this,SLOT(slot_line_edit_change()));
}

PlaylistEditor::~PlaylistEditor()
{
    Debug::debug() << "PlaylistEditor -> wait to finish ..."; 
    m_thread_manager->stopThread();    
    delete m_thread_manager;
    delete model_;
}


void PlaylistEditor::closeEvent ( QCloseEvent * event )
{
    QDialog::closeEvent(event);  
    emit closed();  
}

void PlaylistEditor::hideEvent ( QHideEvent * )
{
    //! force to quit if "esc" is pressed instead of hiding dialog window
    this->setResult(QDialog::Rejected);
    QDialog::reject();
    this->close();    
}
/*******************************************************************************
 ui part
*******************************************************************************/
void PlaylistEditor::create_ui()
{
    //! ui part
    this->setModal(false);     
    this->setWindowTitle(tr("Playlist Editor"));
    this->resize(500,620);

    //In Linux with KDE this code make a window without an close and minimize 
    // and maximize buttons in title bar.
    this->setWindowFlags( Qt::Dialog | Qt::WindowTitleHint );
    
    // set toolbar & action
    QToolBar* toolbar = new QToolBar(this);     
    toolbar->addAction( QIcon(":/images/track-48x48.png"), tr("&Add media to playlist"), this, SLOT(slot_add_file()));
    toolbar->addAction( QIcon(":/images/folder.png"), tr("&Add directory to playlist"), this, SLOT(slot_add_dir()));
    toolbar->addAction( QIcon(":/images/media-url-48x48.png"),tr("&Add Url..."), this, SLOT(slot_add_url()));
    toolbar->addAction( QIcon::fromTheme("edit-clear-list"), tr("Clear playlist"), this, SLOT(slot_clear_playlist()));
    
    ui_action_save_file = new QAction (QIcon(":/images/save-as-32x32.png"), tr("&Save playlist to file"), this);
    ui_action_save_auto = new QAction (QIcon(":/images/save-32x32.png"), tr("&Auto save playlist to YaRock database"), this);
    ui_action_save_file->setEnabled(false);
    ui_action_save_auto->setEnabled(false);
    connect(ui_action_save_file, SIGNAL(triggered()), this, SLOT(slot_save_to_file()));
    connect(ui_action_save_auto, SIGNAL(triggered()), this, SLOT(slot_save_to_db()));
    
    toolbar->addAction( ui_action_save_file );
    toolbar->addAction( ui_action_save_auto );
    
    // set layout
    QGridLayout* gridLayout = new QGridLayout();
    gridLayout->setContentsMargins(4, 0, 0, 0);
    gridLayout->setSpacing(4);

    QLabel* label = new QLabel(tr("name"));
    
    ui_line_edit = new LineEditClear(this);
    ui_line_edit->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Minimum);
    ui_line_edit->clearFocus();
    ui_line_edit->setInactiveText(tr("enter a playlist name"));
    
    gridLayout->addWidget(label, 0, 0, 1, 1);
    gridLayout->addWidget(ui_line_edit, 0, 1, 1, 1);
    
    QFrame* frame = new QFrame();
    frame->setFrameShape(QFrame::StyledPanel);
    frame->setFrameShadow(QFrame::Sunken);
    
    QVBoxLayout *vl0 = new QVBoxLayout(frame);
    vl0->setSpacing(0);
    vl0->setContentsMargins(0, 0, 0, 0);
    vl0->addWidget(view_);
    
    QVBoxLayout *vl1 = new QVBoxLayout(this);
    vl1->setSpacing(0);
    vl1->setContentsMargins(0, 0, 0, 0);
    vl1->addWidget(toolbar);
    vl1->addLayout(gridLayout);
    vl1->addWidget(frame);
        
    //! QDialogButtonBox
    QDialogButtonBox* ui_buttonBox = new QDialogButtonBox();
    ui_buttonBox->setOrientation(Qt::Horizontal);
    ui_buttonBox->setStandardButtons(QDialogButtonBox::Close);
    vl1->addWidget(ui_buttonBox);
    QObject::connect(ui_buttonBox, SIGNAL(rejected()), this, SLOT(slot_on_buttonBox_rejected()));
    
    // avoid focus to be on line edit 
    this->setFocus(Qt::OtherFocusReason);
    
}

void PlaylistEditor::slot_on_buttonBox_rejected()
{
    this->setResult(QDialog::Rejected);
    QDialog::reject();
    this->close();
}

void PlaylistEditor::slot_line_edit_change()
{
    ui_action_save_file->setEnabled( !ui_line_edit->text().isEmpty() );
    ui_action_save_auto->setEnabled( !ui_line_edit->text().isEmpty() );      
}


/*******************************************************************************
 public interface
*******************************************************************************/
void PlaylistEditor::setMediaList(QList<MediaItemPtr> list)
{
    foreach(MediaItemPtr mi, list)   {
      model_->addMediaItem(mi);      
    }  
}


void PlaylistEditor::setPlaylist(MediaItemPtr playlist)
{
    playlist_ = playlist;
    ui_line_edit->setText( playlist->data.at(PLAYLIST::name).toString() );

    model_->clear();
    foreach(MediaItemPtr mi, playlist->childItems)   {
      model_->addMediaItem(mi);      
    }
}


/*******************************************************************************
 Slots
*******************************************************************************/
void PlaylistEditor::slot_clear_playlist()
{
    view_->clearPlaylist();
}

/*******************************************************************************
 Playlist Population 
*******************************************************************************/
void PlaylistEditor::slot_add_file()
{
    // TO DO : repartir du dernier repertoire ouvert 
    QStringList files = QFileDialog::getOpenFileNames(
                this, 
                tr("Add music files or playlist"), 
                QDesktopServices::storageLocation(QDesktopServices::MusicLocation), 
                tr("Audio files(*.mp3 *.wav *.ogg *.flac *.m4a *.aac);;Playlists files(*.m3u *.pls *.xspf);;All files(*.*)"));

    m_thread_manager->playlistAddFiles(files);
}


void PlaylistEditor::slot_add_dir()
{
  // TO DO : repartir du dernier repertoire ouvert 
  QString dirName = QFileDialog::getExistingDirectory(
                      this,
                      tr("Add music directory"),
                      QDesktopServices::storageLocation(QDesktopServices::MusicLocation), 
                      QFileDialog::ShowDirsOnly);
  
    if (!dirName.isEmpty())
      m_thread_manager->playlistAddFile(dirName);  
}


void PlaylistEditor::slot_add_url()
{
    AddStreamDialog stream_dialog(this,false);
      
    if(stream_dialog.exec() == QDialog::Accepted) 
    {
      const QString url   = stream_dialog.url();
        
      if(!QUrl(url).isEmpty() && QUrl(url).isValid()) {
        const QString name  = stream_dialog.name();
        
        MediaItemPtr media = MediaItemPtr(new MediaItem(TYPE_STREAM));
        media->data[STREAM::id]          = -1;
        media->data[STREAM::url]         = url;
        media->data[STREAM::name]        = !name.isEmpty() ? name : url ;
        media->data[STREAM::isFavorite]  = false;
        media->data[STREAM::isPlaying]   = false;
        media->data[STREAM::isBroken]    = false;
        media->data[STREAM::isPlayed]    = false;
        media->data[STREAM::isStopAfter] = false;

        model_->addMediaItem(media);
        media.reset();
      }
      else {
        StatusWidget::instance()->startShortMessage("invalid url can not be added !!", STATUS::TYPE_WARNING, 5000);
      }
    }  
}


/*******************************************************************************
 Playlist Writer 
*******************************************************************************/
void PlaylistEditor::slot_save_to_file()
{
    Debug::debug() << "PlaylistEditor -> slot_save_to_file";

    //! new playlist
    if(playlist_ == MediaItemPtr(0)) 
    {
      QString filename = QFileDialog::getSaveFileName(
                this, 
                tr("Save playlist to file"), 
                QDir::homePath(), 
                tr("m3u playlist (m3u) (*.m3u);;pls playlist (pls) (*.pls);;xspf playlist (xspf) (*.xspf)")
                );

      if (!filename.isEmpty()) 
        m_thread_manager->playlistSaveToFile(filename);   
    }
    //! playlist existante
    else 
    {
      const QString filename = playlist_->data.at(PLAYLIST::url).toString();
      
      m_thread_manager->playlistSaveToFile(filename);
    }
}




void PlaylistEditor::slot_save_to_db()
{
    Debug::debug() << "PlaylistEditor -> slot_save_to_db";
    const QString name = ui_line_edit->text();

    if(name.isEmpty()) 
    {
        QString str = tr("<p>Please, fill the playlist name first</p>");
        QMessageBox message(this);
        message.setWindowTitle(tr("Empty playlist name"));
        message.setText(str.toAscii());
        message.setStandardButtons(QMessageBox::Ok);
        message.resize(445, 120);
        message.exec();      
        return;
    }    
    
    //! new playlist
    if(playlist_ == MediaItemPtr(0)) 
    {
       m_thread_manager->playlistSaveToDb( name );
    }
    //! playlist existante
    else 
    {
       int database_id = playlist_->data.at(PLAYLIST::id).toInt();      
       m_thread_manager->playlistSaveToDb( name , database_id);
    }
    
}
