/* YaRock Player
* Copyright (C) 2010-2013 Sebastien amardeilh (sebastien.amardeilh+yarock@gmail.com)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "StreamView/stream_scene.h"
#include "StreamView/stream_editor.h"

#include "StreamModel/streammodel.h"
#include "StreamModel/icecastmanager.h"
#include "StreamModel/shoutcastmanager.h"
#include "StreamModel/xspfstreamlist.h"

#include "CollectionView/virtualplayqueue.h"
#include "Playlist/playlistmodel.h"

#include "MediaItem/mediaitem.h"
#include "MediaItem/playlist_parser.h"
#include "networkaccess.h"
#include "utilities.h"


#include "button_item.h"

#include "Widget/statuswidget.h"

#include "global_actions.h"
#include "yarocksettings.h"
#include "debug.h"

/*
********************************************************************************
*                                                                              *
*    Class StreamScene                                                         *
*                                                                              *
********************************************************************************
*/
StreamScene::StreamScene(QWidget *parent)
{
    setParent(parent);
    _parentWidget = parent;
    setItemIndexMethod(QGraphicsScene::NoIndex);
    
    _infoSize        = 0;
 
    //! set model  
    _streamModel = StreamModel::instance();
    connect(_streamModel, SIGNAL(signalPlayingStatusChanged()), this, SLOT(update()));

    //! set player  
    _player = EnginePlayer::instance();
  
    //! ice cast
    _iceCastManager = new IceCastManager();
    connect(_iceCastManager, SIGNAL(iceCastUpdated()), this, SIGNAL(repopulateStreamScene())); 
  
    //! shout cast
    _shoutCastManager = new ShoutCastManager();
    connect(_shoutCastManager, SIGNAL(shoutCastUpdated()), this, SIGNAL(repopulateStreamScene())); 
  
    _view_mode             = MAIN::ViewIceCast;
    _currentPlaylist       = "";
    m_mouseGrabbedItem     = 0;
  
    //! item double click 
    QObject::connect(ACTIONS()->value(BROWSER_ITEM_DOUBLE_CLICK), SIGNAL(triggered()), this, SLOT(slot_item_doubleclick()), Qt::DirectConnection);
  
    //! graphic item context menu
    m_graphic_item_menu = new GraphicsItemMenu(0);
    QObject::connect(m_graphic_item_menu, SIGNAL(menu_action_triggered(ENUM_ACTION_ITEM_MENU)), this, SLOT(slot_contextmenu_triggered(ENUM_ACTION_ITEM_MENU)), Qt::DirectConnection);

    //! remove icecast/shoutcast duplicate
    m_remove_duplicate_action = new QAction(QIcon(),QString(),this);
    m_remove_duplicate_action->setCheckable(true);
    m_remove_duplicate_action->setChecked(false);
    QObject::connect(m_remove_duplicate_action, SIGNAL(triggered()), this, SLOT(slot_remove_duplicate_stream()));
}   

StreamScene::~StreamScene()
{
    clear(); 
}   

void StreamScene::setVirtualPlayqueue(VirtualPlayqueue  *queue)
{
    _virtualPlaylist = queue;
}

//! --------------------- playing status method --------------------- 
void StreamScene::updateStreamPlaying(const QString& urlName)
{  
    Debug::debug() << " ---- StreamScene -> updateStreamPlaying" << urlName;
    if(!urlName.isEmpty()) 
    {
      _streamModel->stopPlayingItem();
      if  (!MEDIA::isUrlLocal(urlName)) {
        if(_currentPlaylist.isEmpty())
          _streamModel->updatePlayingItem(urlName);
        else
          _streamModel->updatePlayingItem(_currentPlaylist);
      }
    }
    else
    {  
      _currentPlaylist = "";
      _streamModel->stopPlayingItem();
    }
}
//! --------------------- broken status method --------------------------------- 
void StreamScene::updateStreamBroken(const QString& urlName)
{
    Debug::debug() << " ---- StreamScene -> updateStreamBroken" << urlName;
    
    if(!urlName.isEmpty() && !MEDIA::isUrlLocal(urlName)) 
    {
      if(!_currentPlaylist.isEmpty())
	_streamModel->updateBrokenItem(_currentPlaylist);      
      else
       _streamModel->updateBrokenItem(urlName);
    }
    else if(!_currentPlaylist.isEmpty())
      _streamModel->updateBrokenItem(_currentPlaylist);
}




/*******************************************************************************
    playStream
*******************************************************************************/
void StreamScene::playStream() 
{
    if(!m_mouseGrabbedItem) return;
    StreamGraphicItem *item = static_cast<StreamGraphicItem*>(m_mouseGrabbedItem);
  
    const QString streamUrl = item->media->data.at(STREAM::url).toString();

    /*--------------------------------------------------*/
    /* urls simple                                      */
    /* -------------------------------------------------*/
    if( !MEDIA::isPlaylistFile(streamUrl) && !MEDIA::isShoutCastUrl(streamUrl)) 
    {
      Debug::debug() << " ---- StreamScene -> playStream  we can play it's an url"<< streamUrl;

      _currentPlaylist = "";
      _virtualPlaylist->appendMedia(item->media);
      return;
    }    
    /*--------------------------------------------------*/
    /* playlist remote or shoutcast                     */
    /* -------------------------------------------------*/
    Debug::debug() << " ---- StreamScene -> playStream  it's a playlist -> download to do" ;
    _currentPlaylist = streamUrl;

    //! start asynchronous download
    QObject *reply = HTTP()->get(streamUrl);
    connect(reply, SIGNAL(error(QNetworkReply*)), this, SLOT(updateStreamBroken()));
    connect(reply, SIGNAL(data(QByteArray)), SLOT(slot_streams_fetched(QByteArray)));       
}    

void StreamScene::slot_streams_fetched(QByteArray bytes)
{
    Debug::debug() << " ---- StreamScene -> slot_streams_fetched" ;
  
    if(bytes.isEmpty()) 
    {
      Debug::debug() << " ---- StreamScene -> playStream  received empty bytes";
      StatusWidget::instance()->startShortMessage("receive empty bytes for this stream", STATUS::TYPE_WARNING, 3000);
      updateStreamBroken(_currentPlaylist);
      return;
    }

    QList<MediaItemPtr> list = MEDIA::PlaylistFromBytes(bytes);
    if(list.isEmpty()) 
    {
      Debug::debug() << " ---- StreamScene -> playStream empty playlist downloaded";
      StatusWidget::instance()->startShortMessage("the downloaded playlist is empty", STATUS::TYPE_WARNING, 3000);
      updateStreamBroken(_currentPlaylist);
      return;      
    }

    _virtualPlaylist->appendMedias(list,0);
}


/*******************************************************************************
    enqueueStream
*******************************************************************************/
void StreamScene::enqueueStream(bool is_new_playqueue)
{
    if(!m_mouseGrabbedItem) return;
    
    if(is_new_playqueue)
      PlaylistModel::instance()->clear();
    
    StreamGraphicItem *item = static_cast<StreamGraphicItem*>(m_mouseGrabbedItem);
  
    const QString streamUrl = item->media->data.at(STREAM::url).toString();

    /*--------------------------------------------------*/
    /* urls simple                                      */
    /* -------------------------------------------------*/
    if( !MEDIA::isPlaylistFile(streamUrl) && !MEDIA::isShoutCastUrl(streamUrl)) 
    {
      PlaylistModel::instance()->addMediaItem(item->media);
      return;
    }
    
    /*--------------------------------------------------*/
    /* playlist remote or shoutcast                     */
    /* -------------------------------------------------*/    
    // use thread manager to load streams into playqueue
    m_thread_manager->playlistAddFile(streamUrl);
}
 

/*******************************************************************************
    populateScene
*******************************************************************************/
void StreamScene::populateScene() 
{
    Debug::debug() << "StreamScene -> PopulateScene";
  
    //! clear the scene FIRST
    clear();
  
    switch(_view_mode) {
      case (MAIN::ViewIceCast) : 
        if(!_iceCastManager->isReady()) {
          _iceCastManager->startDownload();
          populateWaitingScene( QObject::tr("Start downloading icecast stream....please wait...") );
          return;
        }
        else if (_iceCastManager->isError()){
          populateWaitingScene( QObject::tr("IceCast Manager failed to download icecast stream !") );
          return;
        }
      
        _streamModel->updateListeStream(_iceCastManager->getIceCastStream());
        populateStreamScene(); 
        break;
    
      case (MAIN::ViewShoutCast) : 
        if(!_shoutCastManager->isReady()) {
          _shoutCastManager->startDownload();
          populateWaitingScene( QObject::tr("Start downloading shoutcast stream....please wait...") );
          return;
        }
        else if (_shoutCastManager->isError()){
          populateWaitingScene( QObject::tr("Shoutcast Manager failed to download shoutcast stream !") );
          return;
        }

        _streamModel->updateListeStream(_shoutCastManager->getShoutCastStream());
        populateStreamScene(); 
        break;
      
      case (MAIN::ViewCustomRadio) :
      {
        XspfStreamList xspf(UTIL::CONFIGDIR + "/radio/" + SETTINGS()->_radiofile);
        xspf.read();

        _streamModel->updateListeStream(xspf.mediaList());
        populateStreamScene();
      }
      break;
      
      case (MAIN::ViewFavoriteRadio) :
      {
        XspfStreamList xspf(UTIL::CONFIGDIR + "/favorite.xspf");
        xspf.read();

        _streamModel->updateListeStream(xspf.mediaList());
        populateFavoriteStreamScene();
      }
      break;

      default: break;
    }
  
    if(_player->engineState() == ENGINE::PLAYING) 
    {
      if( !MEDIA::isUrlLocal (_player->enginePlayingUrl()) )
        updateStreamPlaying(_player->enginePlayingUrl());
    }
    
    //! we need to ajust SceneRect 
    int diff = _parentWidget->height() - itemsBoundingRect().height() + 40;
    int ajust  = ( diff > 0) ? diff : 0; 
    setSceneRect ( itemsBoundingRect().adjusted(-10, -10, +10, +10 + ajust) );
}


/*******************************************************************************
    populateStreamScene
*******************************************************************************/
void StreamScene::populateStreamScene() 
{
    Debug::debug() << " ----StreamScene -> populateStreamScene ----";
  
    int streamRow      = 0;
    int categoriesRow  = 0;
    _infoSize          = 0;
  
    //! categories list (for up/down navigation)
    _listCategorie.clear();
  
    ButtonItem* ui_button  = new ButtonItem();
    ui_button->setText(tr("Reload"));   
    ui_button->setPixmap(QPixmap(":/images/rebuild.png"));
    addItem(ui_button);
    connect(ui_button, SIGNAL(clicked()), this, SLOT(slot_reload_stream_list()));   
    
    ButtonItem* ui_button_2          = new ButtonItem();
    if(_view_mode == MAIN::ViewIceCast || _view_mode == MAIN::ViewShoutCast) 
    {
      if(m_remove_duplicate_action->isChecked())
        ui_button_2->setText(tr("Show duplicate"));
      else
        ui_button_2->setText(tr("Remove duplicate"));

      connect(ui_button_2, SIGNAL(clicked()), m_remove_duplicate_action, SIGNAL(triggered()));   
    }
    else
    {
      ui_button_2->setText(tr("Edit"));
      ui_button_2->setToolTip(tr("Edit stream list"));
      ui_button_2->setPixmap(QPixmap(":/images/edit_info.png"));
      connect(ui_button_2, SIGNAL(clicked()), this, SLOT(slot_edit_stream_list()));   
    }
    
    addItem(ui_button_2);

    int xPos = ui_button->width() + ui_button_2->width() +20+30;
    xPos = (xPos > _parentWidget->width()) ? 30 : (_parentWidget->width() - xPos)/2;

    ui_button->setPos(xPos, 25);
    ui_button_2->setPos(xPos+ ui_button->width() + 20, 25);    


    
    categoriesRow++;
    
    QString currentCategoryName;
    //! stream model loop
    for ( int i = 0; i < _streamModel->itemCount(); i++ ) 
    {
      if(!_streamModel->isStreamFiltered(i)) continue;
    
      if(currentCategoryName != _streamModel->getStreamItem(i)->data.at(STREAM::categorie).toString()) 
      {
        //! new category
        currentCategoryName = _streamModel->getStreamItem(i)->data.at(STREAM::categorie).toString();
     
        // hack pour une présentation correcte on rajoute une fausse ligne de stream avant une nouvelle categorie
        streamRow++; 
        CategorieGraphicItem *cat = new CategorieGraphicItem();
        cat->_catName = currentCategoryName;
        cat->_width   = _parentWidget->width()-20;
        cat->setPos(0 ,categoriesRow*40 + streamRow*20 );

        addItem(cat);
        categoriesRow++;
        _listCategorie.append(cat);
      }

      StreamGraphicItem *stream_item = new StreamGraphicItem();
      stream_item->media = _streamModel->getStreamItem(i);
    
      stream_item->media->data[STREAM::isFavorite] = false;
      stream_item->setPos(30, categoriesRow*40 + streamRow*20);
      stream_item->_width   = _parentWidget->width()-30-20;
      stream_item->setToolTip(stream_item->media->data.at(STREAM::url).toString());
    
      addItem(stream_item); 
 
      streamRow++;
      _infoSize++;
    }
  
    if(_infoSize==0) {
      streamRow++; 
      
      CategorieGraphicItem *cat = new CategorieGraphicItem();
      cat->_catName = tr("Streams");
      cat->_width   = _parentWidget->width()-20;
      cat->setPos( 0 ,categoriesRow*40 + streamRow*20 );

      addItem(cat);
      categoriesRow++;
      
      InfoGraphicItem *item = new InfoGraphicItem();
      item->_text = tr("No entry found");
      item->_width   = _parentWidget->width();
      item->setPos( 0 , 10 + categoriesRow*40 + streamRow*20);
      addItem(item);
    }
}

/*******************************************************************************
    populateFavoriteStreamScene
*******************************************************************************/
void StreamScene::populateFavoriteStreamScene() 
{
    Debug::debug() << " ----StreamScene -> populateStreamScene ----";
  
    int streamRow      = 0;
    int categoriesRow  = 0;
    _infoSize          = 0;
  
    //! categories list (for up/down navigation)
    _listCategorie.clear();
  
    ButtonItem* ui_button  = new ButtonItem();
    ui_button->setText(tr("Reload"));
    ui_button->setPixmap(QPixmap(":/images/rebuild.png"));
    addItem(ui_button);
    connect(ui_button, SIGNAL(clicked()), this, SLOT(slot_reload_stream_list()));

    ButtonItem* ui_button_2          = new ButtonItem();
    ui_button_2->setText(tr("Edit"));
    ui_button_2->setToolTip(tr("Edit stream list"));
    ui_button_2->setPixmap(QPixmap(":/images/edit_info.png"));
    addItem(ui_button_2);      
    connect(ui_button_2, SIGNAL(clicked()), this, SLOT(slot_edit_stream_list()));
      
    int xPos = ui_button->width() + ui_button_2->width() +20+30;
    xPos = (xPos > _parentWidget->width()) ? 30 : (_parentWidget->width() - xPos)/2;

    ui_button->setPos(xPos, 25);
    ui_button_2->setPos(xPos+ ui_button->width() + 20, 25);
    
    
    categoriesRow++;    
    
    QString currentCategoryName;
    //! stream model loop
    for ( int i = 0; i < _streamModel->itemCount(); i++ ) 
    {
      if(!_streamModel->isStreamFiltered(i)) continue;
    
      if(currentCategoryName != _streamModel->getStreamItem(i)->data.at(STREAM::categorie).toString()) 
      {
        //! new category
        currentCategoryName = _streamModel->getStreamItem(i)->data.at(STREAM::categorie).toString();

        // hack pour une présentation correcte on rajoute une fausse ligne de stream avant une nouvelle categorie
        streamRow++; 
      
        CategorieGraphicItem *cat = new CategorieGraphicItem();
        cat->_catName = currentCategoryName;
        cat->_width   = _parentWidget->width()-20;
        cat->setPos( 0 ,categoriesRow*40 + streamRow*20 );

        addItem(cat);
        categoriesRow++;
        _listCategorie.append(cat);
      }

      StreamGraphicItem *stream_item = new StreamGraphicItem();
      stream_item->media = _streamModel->getStreamItem(i);
      stream_item->media->data[STREAM::isFavorite] = true;
      stream_item->_width   = _parentWidget->width()-30-20;
      stream_item->setPos(30, categoriesRow*40 + streamRow*20);
      stream_item->setToolTip(stream_item->media->data.at(STREAM::url).toString());
      addItem(stream_item); 
   
      streamRow++;
      _infoSize++;
    }  
  
    if(_infoSize==0) {
      streamRow++; 
      
      CategorieGraphicItem *cat = new CategorieGraphicItem();
      cat->_catName = tr("Favorite Streams");
      cat->_width   = _parentWidget->width()-20;
      cat->setPos( 0 ,categoriesRow*40 + streamRow*20 );

      addItem(cat);
      categoriesRow++;
      
      InfoGraphicItem *item = new InfoGraphicItem();
      item->_text = tr("No entry found");
      item->_width   = _parentWidget->width();
      item->setPos( 0 , 10 + categoriesRow*40 + streamRow*20);
      addItem(item);
    }
}

void StreamScene::populateWaitingScene(QString text) 
{
    _listCategorie.clear();
  
    CategorieGraphicItem *cat = new CategorieGraphicItem();
    cat->_catName = tr("Streams");
    cat->_width   = _parentWidget->width()-20;
    cat->setPos( 0 ,30);

    addItem(cat);
      
    InfoGraphicItem *item = new InfoGraphicItem();
    item->_text = text;
    item->_width   = _parentWidget->width();
    item->setPos( 0 , 30 + 40);
    addItem(item);
}


/*******************************************************************************
    contextMenuEvent
*******************************************************************************/
void StreamScene::contextMenuEvent(QGraphicsSceneContextMenuEvent *event) 
{
    m_mouseGrabbedItem = this->mouseGrabberItem();
    if(!m_mouseGrabbedItem) return;
  
    m_graphic_item_menu->setGraphicItem(m_mouseGrabbedItem);
    m_graphic_item_menu->exec(event->screenPos());
}

void StreamScene::slot_contextmenu_triggered(ENUM_ACTION_ITEM_MENU id)
{
    switch(id)
    {
      case STREAM_PLAY         : playStream();            break;
      case STREAM_EDIT         : break;
      case STREAM_QUEUE_END    : enqueueStream(false);    break;
      case STREAM_QUEUE_NEW    : enqueueStream(true);     break;
      case STREAM_FAVORITE     : updateStreamFavorite();  break;
      default:break;
    }
}
   

/*******************************************************************************
    slot_item_doubleclick
*******************************************************************************/
void StreamScene::slot_item_doubleclick()
{
   QList<QGraphicsView *> view = this->views();
   if(view.isEmpty()) {
     //Debug::debug() << " ---- StreamScene:: View Not used";    
     return;
   }
  
   //Debug::debug() << " ---- StreamScene::streamDoubleClick";
   QMap<ENUM_ACTION, QAction*> *actions = ACTIONS();
   
   m_mouseGrabbedItem = qvariant_cast<QGraphicsItem *>( (actions->value(BROWSER_ITEM_DOUBLE_CLICK))->data() );

   if( !m_mouseGrabbedItem || (m_mouseGrabbedItem->type() != GraphicsItem::StreamType) )
     return; 
   
   if(Qt::ControlModifier == QApplication::keyboardModifiers()) 
   {
     enqueueStream(false);
   }
   else 
   {
     playStream();
   }   
}

/*******************************************************************************
    Manage favorite
*******************************************************************************/
void StreamScene::updateStreamFavorite() 
{
    Debug::debug() << " ---- StreamScene -> updateStreamFavorite";

    if(!m_mouseGrabbedItem) return;
    StreamGraphicItem *item = static_cast<StreamGraphicItem*>(m_mouseGrabbedItem);
    
    bool isFavorite   = _streamModel->itemIsFavorite(item->media);

    Debug::debug() << " ---- StreamScene -> updateStreamFavorite isFavorite" << isFavorite;
    
    XspfStreamList xspf_favorite(UTIL::CONFIGDIR + "/favorite.xspf");
    xspf_favorite.read();
    
    if(isFavorite)  // REMOVE IT
    {
      bool removed = false;
      QList<MediaItemPtr> currentList = xspf_favorite.mediaList();
      
      // remove stream if exist 
      foreach (MediaItemPtr mi, currentList) {
        if(mi->data[STREAM::url].toString() == item->media->data[STREAM::url].toString()) {
          currentList.removeAt(currentList.indexOf(mi));
          removed = true;
          break;
        }
      }

      if(removed) {
        qSort(currentList.begin(), currentList.end(), MEDIA::compareStreamCategorie);
        
        xspf_favorite.setMediaList(currentList);
        xspf_favorite.write();

        populateScene();
      }  
    }
    else // ADD IT
    {
      bool exist = false;
      // get current favorite
      QList<MediaItemPtr> currentList = xspf_favorite.mediaList();

      // add stream if does not exist 
      foreach (MediaItemPtr mi, currentList) {
        if(mi->data.at(STREAM::url).toString() == item->media->data.at(STREAM::url).toString()) {
          exist = true;
          break;
        }
      }
  
      if(!exist) {
        currentList.append( item->media );
        qSort(currentList.begin(), currentList.end(), MEDIA::compareStreamCategorie);

        xspf_favorite.setMediaList(currentList);
        xspf_favorite.write();

        populateScene();
      }      
    }
}



//! -------------------- slot_reload_stream_list --------------------------------
void StreamScene::slot_reload_stream_list()
{
    if(_view_mode == MAIN::ViewIceCast) {
       QFile file(UTIL::CONFIGDIR + "/radio/" + "icecast.xml");
       if(file.exists() && _iceCastManager->isReady()) file.remove();
    }
    else if(_view_mode == MAIN::ViewShoutCast) {
       QFile file(UTIL::CONFIGDIR + "/radio/" + "shoutcast.xml");
       if(file.exists() && _shoutCastManager->isReady()) file.remove();
    }

    populateScene();
}


//! --------------------- slot_check_stream -------------------------------------
void StreamScene::slot_check_stream()
{
    for ( int i = 0; i < _streamModel->itemCount(); i++ ) 
    {
      Debug::debug() << " StreamScene::slot_check_stream";
      
      MediaItemPtr mi = _streamModel->getStreamItem(i);
      QString url = mi->data.at(STREAM::url).toString();
      bool isUrlOK = HTTP()->syncTest( QUrl(url) );
      mi->data[STREAM::isBroken] = !isUrlOK;
    }
}


//! --------------------- slot_remove_duplicate_stream -------------------------
void StreamScene::slot_remove_duplicate_stream()
{
    Debug::debug() << " StreamScene::slot_remove_duplicate_stream";

    bool isChecked = m_remove_duplicate_action->isChecked();
    m_remove_duplicate_action->setChecked( !isChecked );
    
    _iceCastManager->setDuplicate(isChecked);
    _shoutCastManager->setDuplicate(isChecked);
     
    populateScene();
}



//! --------------------- slot_edit_stream_list --------------------------------
void StreamScene::slot_edit_stream_list()
{
    QString file_path;
    
    if(_view_mode == MAIN::ViewCustomRadio) 
    {
      file_path = UTIL::CONFIGDIR + "/radio/" + SETTINGS()->_radiofile;
    }
    else  if(_view_mode == MAIN::ViewFavoriteRadio) 
    {
      file_path = UTIL::CONFIGDIR + "/favorite.xspf";
    }
    Debug::debug() << " StreamScene::slot_edit_stream_list -> file_path : " << file_path;
    
    StreamEditor* stream_editor = new StreamEditor(0);
    connect(stream_editor, SIGNAL(stream_list_saved()), this, SLOT(populateScene()));
    stream_editor->load_stream_file(file_path);
    
    stream_editor->show();
}

